"""Defines the interface of a cachier caching core."""

# This file is part of Cachier.
# https://github.com/python-cachier/cachier

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2016, Shay Palachy <shaypal5@gmail.com>

import abc  # for the _BaseCore abstract base class
import inspect
import threading
from typing import Callable, Optional, Tuple

from .._types import HashFunc
from ..config import CacheEntry, _update_with_defaults


class RecalculationNeeded(Exception):
    """Exception raised when a recalculation is needed."""

    pass


def _get_func_str(func: Callable) -> str:
    return f".{func.__module__}.{func.__name__}"


class _BaseCore:
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        hash_func: Optional[HashFunc],
        wait_for_calc_timeout: Optional[int],
    ):
        self.hash_func = _update_with_defaults(hash_func, "hash_func")
        self.wait_for_calc_timeout = wait_for_calc_timeout
        self.lock = threading.RLock()

    def set_func(self, func):
        """Set the function this core will use.

        This has to be set before any method is called. Also determine if the
        function is an object method.

        """
        # unwrap if the function is functools.partial
        if hasattr(func, "func"):
            func = func.func
        func_params = list(inspect.signature(func).parameters)
        self.func_is_method = func_params and func_params[0] == "self"
        self.func = func

    def get_key(self, args, kwds):
        """Return a unique key based on the arguments provided."""
        return self.hash_func(args, kwds)

    def get_entry(self, args, kwds) -> Tuple[str, Optional[CacheEntry]]:
        """Get entry based on given arguments.

        Return the result mapped to the given arguments in this core's cache,
        if such a mapping exists.

        """
        key = self.get_key(args, kwds)
        return self.get_entry_by_key(key)

    def precache_value(self, args, kwds, value_to_cache):
        """Write a precomputed value into the cache."""
        key = self.get_key(args, kwds)
        self.set_entry(key, value_to_cache)
        return value_to_cache

    def check_calc_timeout(self, time_spent):
        """Raise an exception if a recalculation is needed."""
        calc_timeout = _update_with_defaults(
            self.wait_for_calc_timeout, "wait_for_calc_timeout"
        )
        if calc_timeout > 0 and (time_spent >= calc_timeout):
            raise RecalculationNeeded()

    @abc.abstractmethod
    def get_entry_by_key(self, key: str) -> Tuple[str, Optional[CacheEntry]]:
        """Get entry based on given key.

        Return the result mapped to the given key in this core's cache, if such
        a mapping exists.

        """

    @abc.abstractmethod
    def set_entry(self, key: str, func_res):
        """Map the given result to the given key in this core's cache."""

    @abc.abstractmethod
    def mark_entry_being_calculated(self, key: str) -> None:
        """Mark the entry mapped by the given key as being calculated."""

    @abc.abstractmethod
    def mark_entry_not_calculated(self, key: str) -> None:
        """Mark the entry mapped by the given key as not being calculated."""

    @abc.abstractmethod
    def wait_on_entry_calc(self, key: str) -> None:
        """Wait on the entry with keys being calculated and returns result."""

    @abc.abstractmethod
    def clear_cache(self) -> None:
        """Clear the cache of this core."""

    @abc.abstractmethod
    def clear_being_calculated(self) -> None:
        """Mark all entries in this cache as not being calculated."""



"""A command-line interface for cachier."""

import click

from cachier.core import _set_max_workers


@click.group()
def cli():
    """A command-line interface for cachier."""  # noqa: D401


@cli.command("Limits the number of worker threads used by cachier.")
@click.argument("max_workers", nargs=1, type=int)
def set_max_workers(max_workers):
    """Limits the number of worker threads used by cachier."""
    _set_max_workers(max_workers)



from ._version import __version__
from .config import (
    disable_caching,
    enable_caching,
    get_default_params,
    get_global_params,
    set_default_params,
    set_global_params,
)
from .core import cachier

__all__ = [
    "cachier",
    "set_default_params",
    "get_default_params",
    "set_global_params",
    "get_global_params",
    "enable_caching",
    "disable_caching",
    "__version__",
]



"""A memory-based caching core for cachier."""

import threading
from datetime import datetime
from typing import Any, Dict, Optional, Tuple

from .._types import HashFunc
from ..config import CacheEntry
from .base import _BaseCore, _get_func_str


class _MemoryCore(_BaseCore):
    """The memory core class for cachier."""

    def __init__(
        self,
        hash_func: Optional[HashFunc],
        wait_for_calc_timeout: Optional[int],
    ):
        super().__init__(hash_func, wait_for_calc_timeout)
        self.cache: Dict[str, CacheEntry] = {}

    def _hash_func_key(self, key: str) -> str:
        return f"{_get_func_str(self.func)}:{key}"

    def get_entry_by_key(
        self, key: str, reload=False
    ) -> Tuple[str, Optional[CacheEntry]]:
        with self.lock:
            return key, self.cache.get(self._hash_func_key(key), None)

    def set_entry(self, key: str, func_res: Any) -> None:
        hash_key = self._hash_func_key(key)
        with self.lock:
            try:
                # we need to retain the existing condition so that
                # mark_entry_not_calculated can notify all possibly-waiting
                # threads about it
                cond = self.cache[hash_key]._condition
            except KeyError:  # pragma: no cover
                cond = None
            self.cache[hash_key] = CacheEntry(
                value=func_res,
                time=datetime.now(),
                stale=False,
                _processing=False,
                _condition=cond,
                _completed=True,
            )

    def mark_entry_being_calculated(self, key: str) -> None:
        with self.lock:
            condition = threading.Condition()
            hash_key = self._hash_func_key(key)
            if hash_key in self.cache:
                self.cache[hash_key]._processing = True
                self.cache[hash_key]._condition = condition
            # condition.acquire()
            else:
                self.cache[hash_key] = CacheEntry(
                    value=None,
                    time=datetime.now(),
                    stale=False,
                    _processing=True,
                    _condition=condition,
                )

    def mark_entry_not_calculated(self, key: str) -> None:
        hash_key = self._hash_func_key(key)
        with self.lock:
            if hash_key not in self.cache:
                return  # that's ok, we don't need an entry in that case
            entry = self.cache[hash_key]
            entry._processing = False
            cond = entry._condition
            if cond:
                cond.acquire()
                cond.notify_all()
                cond.release()
                entry._condition = None

    def wait_on_entry_calc(self, key: str) -> Any:
        hash_key = self._hash_func_key(key)
        with self.lock:  # pragma: no cover
            entry = self.cache[hash_key]
            if entry is None:
                return None
            if not entry._processing:
                return entry.value
        if entry._condition is None:
            raise RuntimeError("No condition set for entry")
        entry._condition.acquire()
        entry._condition.wait()
        entry._condition.release()
        return self.cache[hash_key].value

    def clear_cache(self) -> None:
        with self.lock:
            self.cache.clear()

    def clear_being_calculated(self) -> None:
        with self.lock:
            for entry in self.cache.values():
                entry._processing = False
                entry._condition = None



"""A SQLAlchemy-based caching core for cachier."""

import pickle
import threading
from datetime import datetime
from typing import Any, Callable, Optional, Tuple, Union

try:
    from sqlalchemy import (
        Boolean,
        Column,
        DateTime,
        Index,
        LargeBinary,
        String,
        and_,
        create_engine,
        delete,
        insert,
        select,
        update,
    )
    from sqlalchemy.engine import Engine
    from sqlalchemy.orm import declarative_base, sessionmaker

    SQLALCHEMY_AVAILABLE = True
except ImportError:
    SQLALCHEMY_AVAILABLE = False

from .._types import HashFunc
from ..config import CacheEntry
from .base import RecalculationNeeded, _BaseCore, _get_func_str

if SQLALCHEMY_AVAILABLE:
    Base = declarative_base()

    class CacheTable(Base):  # type: ignore[misc, valid-type]
        """SQLAlchemy model for cachier cache entries."""

        __tablename__ = "cachier_cache"
        id = Column(String, primary_key=True)
        function_id = Column(String, index=True, nullable=False)
        key = Column(String, index=True, nullable=False)
        value = Column(LargeBinary, nullable=True)
        timestamp = Column(DateTime, nullable=False)
        stale = Column(Boolean, default=False)
        processing = Column(Boolean, default=False)
        completed = Column(Boolean, default=False)
        __table_args__ = (
            Index("ix_func_key", "function_id", "key", unique=True),
        )


class _SQLCore(_BaseCore):
    """SQLAlchemy-based core for Cachier, supporting SQL-based backends.

    This should work with SQLite, PostgreSQL and so on.

    """

    def __init__(
        self,
        hash_func: Optional[HashFunc],
        sql_engine: Optional[Union[str, "Engine", Callable[[], "Engine"]]],
        wait_for_calc_timeout: Optional[int] = None,
    ):
        if not SQLALCHEMY_AVAILABLE:
            raise ImportError(
                "SQLAlchemy is required for the SQL core. "
                "Install with `pip install SQLAlchemy`."
            )
        super().__init__(
            hash_func=hash_func, wait_for_calc_timeout=wait_for_calc_timeout
        )
        self._engine = self._resolve_engine(sql_engine)
        self._Session = sessionmaker(bind=self._engine)
        Base.metadata.create_all(self._engine)
        self._lock = threading.RLock()
        self._func_str = None

    def _resolve_engine(self, sql_engine):
        if isinstance(sql_engine, Engine):
            return sql_engine
        if isinstance(sql_engine, str):
            return create_engine(sql_engine, future=True)
        if callable(sql_engine):
            return sql_engine()
        raise ValueError(
            "sql_engine must be a SQLAlchemy Engine, connection string, "
            "or callable returning an Engine."
        )

    def set_func(self, func):
        super().set_func(func)
        self._func_str = _get_func_str(func)

    def get_entry_by_key(self, key: str) -> Tuple[str, Optional[CacheEntry]]:
        with self._lock, self._Session() as session:
            row = session.execute(
                select(CacheTable).where(
                    and_(
                        CacheTable.function_id == self._func_str,
                        CacheTable.key == key,
                    )
                )
            ).scalar_one_or_none()
            if not row:
                return key, None
            value = pickle.loads(row.value) if row.value is not None else None
            entry = CacheEntry(
                value=value,
                time=row.timestamp,
                stale=row.stale,
                _processing=row.processing,
                _completed=row.completed,
            )
            return key, entry

    def set_entry(self, key: str, func_res: Any) -> None:
        with self._lock, self._Session() as session:
            thebytes = pickle.dumps(func_res)
            now = datetime.now()
            base_insert = insert(CacheTable)
            stmt = (
                base_insert.values(
                    id=f"{self._func_str}:{key}",
                    function_id=self._func_str,
                    key=key,
                    value=thebytes,
                    timestamp=now,
                    stale=False,
                    processing=False,
                    completed=True,
                ).on_conflict_do_update(
                    index_elements=[CacheTable.function_id, CacheTable.key],
                    set_={
                        "value": thebytes,
                        "timestamp": now,
                        "stale": False,
                        "processing": False,
                        "completed": True,
                    },
                )
                if hasattr(base_insert, "on_conflict_do_update")
                else None
            )
            # Fallback for non-SQLite/Postgres: try update, else insert
            if stmt:
                session.execute(stmt)
            else:
                row = session.execute(
                    select(CacheTable).where(
                        and_(
                            CacheTable.function_id == self._func_str,
                            CacheTable.key == key,
                        )
                    )
                ).scalar_one_or_none()
                if row:
                    session.execute(
                        update(CacheTable)
                        .where(
                            and_(
                                CacheTable.function_id == self._func_str,
                                CacheTable.key == key,
                            )
                        )
                        .values(
                            value=thebytes,
                            timestamp=now,
                            stale=False,
                            processing=False,
                            completed=True,
                        )
                    )
                else:
                    session.add(
                        CacheTable(
                            id=f"{self._func_str}:{key}",
                            function_id=self._func_str,
                            key=key,
                            value=thebytes,
                            timestamp=now,
                            stale=False,
                            processing=False,
                            completed=True,
                        )
                    )
            session.commit()

    def mark_entry_being_calculated(self, key: str) -> None:
        with self._lock, self._Session() as session:
            row = session.execute(
                select(CacheTable).where(
                    and_(
                        CacheTable.function_id == self._func_str,
                        CacheTable.key == key,
                    )
                )
            ).scalar_one_or_none()
            if row:
                session.execute(
                    update(CacheTable)
                    .where(
                        and_(
                            CacheTable.function_id == self._func_str,
                            CacheTable.key == key,
                        )
                    )
                    .values(processing=True)
                )
            else:
                session.add(
                    CacheTable(
                        id=f"{self._func_str}:{key}",
                        function_id=self._func_str,
                        key=key,
                        value=None,
                        timestamp=datetime.now(),
                        stale=False,
                        processing=True,
                        completed=False,
                    )
                )
            session.commit()

    def mark_entry_not_calculated(self, key: str) -> None:
        with self._lock, self._Session() as session:
            session.execute(
                update(CacheTable)
                .where(
                    and_(
                        CacheTable.function_id == self._func_str,
                        CacheTable.key == key,
                    )
                )
                .values(processing=False)
            )
            session.commit()

    def wait_on_entry_calc(self, key: str) -> Any:
        import time

        time_spent = 0
        while True:
            with self._lock, self._Session() as session:
                row = session.execute(
                    select(CacheTable).where(
                        and_(
                            CacheTable.function_id == self._func_str,
                            CacheTable.key == key,
                        )
                    )
                ).scalar_one_or_none()
                if not row:
                    raise RecalculationNeeded()
                if not row.processing:
                    return (
                        pickle.loads(row.value)
                        if row.value is not None
                        else None
                    )
            time.sleep(1)
            time_spent += 1
            self.check_calc_timeout(time_spent)

    def clear_cache(self) -> None:
        with self._lock, self._Session() as session:
            session.execute(
                delete(CacheTable).where(
                    CacheTable.function_id == self._func_str
                )
            )
            session.commit()

    def clear_being_calculated(self) -> None:
        with self._lock, self._Session() as session:
            session.execute(
                update(CacheTable)
                .where(
                    and_(
                        CacheTable.function_id == self._func_str,
                        CacheTable.processing,
                    )
                )
                .values(processing=False)
            )
            session.commit()



# This file is part of Cachier.
# https://github.com/python-cachier/cachier

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2024, Jirka Borovec <***@gmail.com>

import os

_PATH_HERE = os.path.dirname(__file__)
_PATH_VERSION = os.path.join(_PATH_HERE, "version.info")
_RELEASING_PROCESS = os.getenv("RELEASING_PROCESS", "0") == "1"

with open(_PATH_VERSION) as fopen:
    __version__ = fopen.read().strip()


def _get_git_sha() -> str:
    from subprocess import DEVNULL, check_output

    args = ["git", "rev-parse", "--short", "HEAD"]
    out = check_output(args, stderr=DEVNULL)  # noqa: S603
    return out.decode("utf-8").strip()


if not _RELEASING_PROCESS:
    try:
        sha_short = _get_git_sha()
        # print(f"Version enriched with git commit hash: {__version__}.")
    except Exception:
        # print("Failed to get the git commit hash,"
        #       f" falling back to base version {__version__}.")
        sha_short = ""
    __version__ += f".dev+{sha_short}" if sha_short else ".dev"


__all__ = ["__version__"]



from typing import TYPE_CHECKING, Callable, Literal

if TYPE_CHECKING:
    import pymongo.collection


HashFunc = Callable[..., str]
Mongetter = Callable[[], "pymongo.collection.Collection"]
Backend = Literal["pickle", "mongo", "memory"]



"""A pickle-based caching core for cachier."""

# This file is part of Cachier.
# https://github.com/python-cachier/cachier

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2016, Shay Palachy <shaypal5@gmail.com>
import os
import pickle  # for local caching
from datetime import datetime
from typing import Any, Dict, Optional, Tuple, Union

import portalocker  # to lock on pickle cache IO
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

from .._types import HashFunc
from ..config import CacheEntry, _update_with_defaults

# Alternative:  https://github.com/WoLpH/portalocker
from .base import _BaseCore


class _PickleCore(_BaseCore):
    """The pickle core class for cachier."""

    class CacheChangeHandler(PatternMatchingEventHandler):
        """Handles cache-file modification events."""

        def __init__(self, filename, core, key):
            PatternMatchingEventHandler.__init__(
                self,
                patterns=["*" + filename],
                ignore_patterns=None,
                ignore_directories=True,
                case_sensitive=False,
            )
            self.core = core
            self.key = key
            self.observer = None
            self.value = None

        def inject_observer(self, observer) -> None:
            """Inject the observer running this handler."""
            self.observer = observer

        def _check_calculation(self) -> None:
            entry = self.core.get_entry_by_key(self.key, True)[1]
            try:
                if not entry._processing:
                    # print('stopping observer!')
                    self.value = entry.value
                    self.observer.stop()
                # else:
                #     print('NOT stopping observer... :(')
            except AttributeError:  # catching entry being None
                self.value = None
                self.observer.stop()

        def on_created(self, event) -> None:
            """A Watchdog Event Handler method."""  # noqa: D401
            self._check_calculation()  # pragma: no cover

        def on_modified(self, event) -> None:
            """A Watchdog Event Handler method."""  # noqa: D401
            self._check_calculation()

    def __init__(
        self,
        hash_func: Optional[HashFunc],
        pickle_reload: Optional[bool],
        cache_dir: Optional[Union[str, os.PathLike]],
        separate_files: Optional[bool],
        wait_for_calc_timeout: Optional[int],
    ):
        super().__init__(hash_func, wait_for_calc_timeout)
        self._cache_dict: Dict[str, CacheEntry] = {}
        self.reload = _update_with_defaults(pickle_reload, "pickle_reload")
        self.cache_dir = os.path.expanduser(
            _update_with_defaults(cache_dir, "cache_dir")
        )
        self.separate_files = _update_with_defaults(
            separate_files, "separate_files"
        )
        self._cache_used_fpath = ""

    @property
    def cache_fname(self) -> str:
        fname = f".{self.func.__module__}.{self.func.__qualname__}"
        return fname.replace("<", "_").replace(">", "_")

    @property
    def cache_fpath(self) -> str:
        os.makedirs(self.cache_dir, exist_ok=True)
        return os.path.abspath(
            os.path.join(os.path.realpath(self.cache_dir), self.cache_fname)
        )

    @staticmethod
    def _convert_legacy_cache_entry(
        entry: Union[dict, CacheEntry],
    ) -> CacheEntry:
        if isinstance(entry, CacheEntry):
            return entry
        return CacheEntry(
            value=entry["value"],
            time=entry["time"],
            stale=entry["stale"],
            _processing=entry["being_calculated"],
            _condition=entry.get("condition", None),
        )

    def _load_cache_dict(self) -> Dict[str, CacheEntry]:
        try:
            with portalocker.Lock(self.cache_fpath, mode="rb") as cf:
                cache = pickle.load(cf)  # noqa: S301
            self._cache_used_fpath = str(self.cache_fpath)
        except (FileNotFoundError, EOFError):
            cache = {}
        return {
            k: _PickleCore._convert_legacy_cache_entry(v)
            for k, v in cache.items()
        }

    def get_cache_dict(self, reload: bool = False) -> Dict[str, CacheEntry]:
        if self._cache_used_fpath != self.cache_fpath:
            # force reload if the cache file has changed
            # this change is dies to using different wrapped function
            reload = True
        if self._cache_dict and not (self.reload or reload):
            return self._cache_dict
        with self.lock:
            self._cache_dict = self._load_cache_dict()
        return self._cache_dict

    def _load_cache_by_key(
        self, key=None, hash_str=None
    ) -> Optional[CacheEntry]:
        fpath = self.cache_fpath
        fpath += f"_{hash_str or key}"
        try:
            with portalocker.Lock(fpath, mode="rb") as cache_file:
                entry = pickle.load(cache_file)  # noqa: S301
            return _PickleCore._convert_legacy_cache_entry(entry)
        except (FileNotFoundError, EOFError):
            return None

    def _clear_all_cache_files(self) -> None:
        path, name = os.path.split(self.cache_fpath)
        for subpath in os.listdir(path):
            if subpath.startswith(f"{name}_"):
                os.remove(os.path.join(path, subpath))

    def _clear_being_calculated_all_cache_files(self) -> None:
        path, name = os.path.split(self.cache_fpath)
        for subpath in os.listdir(path):
            if subpath.startswith(name):
                entry = self._load_cache_by_key(
                    hash_str=subpath.split("_")[-1]
                )
                if entry is not None:
                    entry._processing = False
                    self._save_cache(entry, hash_str=subpath.split("_")[-1])

    def _save_cache(
        self,
        cache: Union[Dict[str, CacheEntry], CacheEntry],
        separate_file_key: Optional[str] = None,
        hash_str: Optional[str] = None,
    ) -> None:
        if separate_file_key and not isinstance(cache, CacheEntry):
            raise ValueError(
                "`separate_file_key` should only be used with a CacheEntry"
            )
        fpath = self.cache_fpath
        if separate_file_key is not None:
            fpath += f"_{separate_file_key}"
        elif hash_str is not None:
            fpath += f"_{hash_str}"
        with self.lock:
            with portalocker.Lock(fpath, mode="wb") as cf:
                pickle.dump(cache, cf, protocol=4)
            # the same as check for separate_file, but changed for typing
            if isinstance(cache, dict):
                self._cache_dict = cache
                self._cache_used_fpath = str(self.cache_fpath)

    def get_entry_by_key(
        self, key: str, reload: bool = False
    ) -> Tuple[str, Optional[CacheEntry]]:
        if self.separate_files:
            return key, self._load_cache_by_key(key)
        return key, self.get_cache_dict(reload).get(key)

    def set_entry(self, key: str, func_res: Any) -> None:
        key_data = CacheEntry(
            value=func_res,
            time=datetime.now(),
            stale=False,
            _processing=False,
            _completed=True,
        )
        if self.separate_files:
            self._save_cache(key_data, key)
            return  # pragma: no cover

        with self.lock:
            cache = self.get_cache_dict()
            cache[key] = key_data
            self._save_cache(cache)

    def mark_entry_being_calculated_separate_files(self, key: str) -> None:
        self._save_cache(
            CacheEntry(
                value=None,
                time=datetime.now(),
                stale=False,
                _processing=True,
            ),
            separate_file_key=key,
        )

    def _mark_entry_not_calculated_separate_files(self, key: str) -> None:
        _, entry = self.get_entry_by_key(key)
        if entry is None:
            return  # that's ok, we don't need an entry in that case
        entry._processing = False
        self._save_cache(entry, separate_file_key=key)

    def mark_entry_being_calculated(self, key: str) -> None:
        if self.separate_files:
            self.mark_entry_being_calculated_separate_files(key)
            return  # pragma: no cover

        with self.lock:
            cache = self.get_cache_dict()
            if key in cache:
                cache[key]._processing = True
            else:
                cache[key] = CacheEntry(
                    value=None,
                    time=datetime.now(),
                    stale=False,
                    _processing=True,
                )
            self._save_cache(cache)

    def mark_entry_not_calculated(self, key: str) -> None:
        if self.separate_files:
            self._mark_entry_not_calculated_separate_files(key)
        with self.lock:
            cache = self.get_cache_dict()
            # that's ok, we don't need an entry in that case
            if isinstance(cache, dict) and key in cache:
                cache[key]._processing = False
                self._save_cache(cache)

    def wait_on_entry_calc(self, key: str) -> Any:
        if self.separate_files:
            entry = self._load_cache_by_key(key)
            filename = f"{self.cache_fname}_{key}"
        else:
            with self.lock:
                entry = self.get_cache_dict()[key]
            filename = self.cache_fname
        if entry and not entry._processing:
            return entry.value
        event_handler = _PickleCore.CacheChangeHandler(
            filename=filename, core=self, key=key
        )
        observer = Observer()
        event_handler.inject_observer(observer)
        observer.schedule(event_handler, path=self.cache_dir, recursive=True)
        observer.start()
        time_spent = 0
        while observer.is_alive():
            observer.join(timeout=1.0)
            time_spent += 1
            self.check_calc_timeout(time_spent)
        return event_handler.value

    def clear_cache(self) -> None:
        if self.separate_files:
            self._clear_all_cache_files()
        else:
            self._save_cache({})

    def clear_being_calculated(self) -> None:
        if self.separate_files:
            self._clear_being_calculated_all_cache_files()
            return  # pragma: no cover

        with self.lock:
            cache = self.get_cache_dict()
            for key in cache:
                cache[key]._processing = False
            self._save_cache(cache)



"""Persistent, stale-free memoization decorators for Python."""

# This file is part of Cachier.
# https://github.com/python-cachier/cachier

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2016, Shay Palachy <shaypal5@gmail.com>

import inspect
import os
import warnings
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from functools import wraps
from typing import Any, Callable, Optional, Union
from warnings import warn

from .config import (
    Backend,
    HashFunc,
    Mongetter,
    _update_with_defaults,
)
from .cores.base import RecalculationNeeded, _BaseCore
from .cores.memory import _MemoryCore
from .cores.mongo import _MongoCore
from .cores.pickle import _PickleCore
from .cores.sql import _SQLCore

MAX_WORKERS_ENVAR_NAME = "CACHIER_MAX_WORKERS"
DEFAULT_MAX_WORKERS = 8


def _max_workers():
    return int(os.environ.get(MAX_WORKERS_ENVAR_NAME, DEFAULT_MAX_WORKERS))


def _set_max_workers(max_workers):
    os.environ[MAX_WORKERS_ENVAR_NAME] = str(max_workers)
    _get_executor(True)


def _get_executor(reset=False):
    if reset or not hasattr(_get_executor, "executor"):
        _get_executor.executor = ThreadPoolExecutor(_max_workers())
    return _get_executor.executor


def _function_thread(core, key, func, args, kwds):
    try:
        func_res = func(*args, **kwds)
        core.set_entry(key, func_res)
    except BaseException as exc:
        print(f"Function call failed with the following exception:\n{exc}")


def _calc_entry(core, key, func, args, kwds) -> Optional[Any]:
    core.mark_entry_being_calculated(key)
    try:
        func_res = func(*args, **kwds)
        core.set_entry(key, func_res)
        return func_res
    finally:
        core.mark_entry_not_calculated(key)


def _convert_args_kwargs(
    func, _is_method: bool, args: tuple, kwds: dict
) -> dict:
    """Convert mix of positional and keyword arguments to aggregated kwargs."""
    # unwrap if the function is functools.partial
    if hasattr(func, "func"):
        args = func.args + args
        kwds.update({k: v for k, v in func.keywords.items() if k not in kwds})
        func = func.func
    func_params = list(inspect.signature(func).parameters)
    args_as_kw = dict(
        zip(func_params[1:], args[1:])
        if _is_method
        else zip(func_params, args)
    )
    # init with default values
    kwargs = {
        k: v.default
        for k, v in inspect.signature(func).parameters.items()
        if v.default is not inspect.Parameter.empty
    }
    # merge args expanded as kwargs and the original kwds
    kwargs.update(dict(**args_as_kw, **kwds))
    return OrderedDict(sorted(kwargs.items()))


def _pop_kwds_with_deprecation(kwds, name: str, default_value: bool):
    if name in kwds:
        warnings.warn(
            f"`{name}` is deprecated and will be removed in a future release,"
            " use `cachier__` alternative instead.",
            DeprecationWarning,
            stacklevel=2,
        )
    return kwds.pop(name, default_value)


def cachier(
    hash_func: Optional[HashFunc] = None,
    hash_params: Optional[HashFunc] = None,
    backend: Optional[Backend] = None,
    mongetter: Optional[Mongetter] = None,
    sql_engine: Optional[Union[str, Any, Callable[[], Any]]] = None,
    stale_after: Optional[timedelta] = None,
    next_time: Optional[bool] = None,
    cache_dir: Optional[Union[str, os.PathLike]] = None,
    pickle_reload: Optional[bool] = None,
    separate_files: Optional[bool] = None,
    wait_for_calc_timeout: Optional[int] = None,
    allow_none: Optional[bool] = None,
):
    """Wrap as a persistent, stale-free memoization decorator.

    The positional and keyword arguments to the wrapped function must be
    hashable (i.e. Python's immutable built-in objects, not mutable
    containers). Also, notice that since objects which are instances of
    user-defined classes are hashable but all compare unequal (their hash
    value is their id), equal objects across different sessions will not yield
    identical keys.

    Arguments:
    ---------
    hash_func : callable, optional
        A callable that gets the args and kwargs from the decorated function
        and returns a hash key for them. This parameter can be used to enable
        the use of cachier with functions that get arguments that are not
        automatically hashable by Python.
    hash_params : callable, optional
    backend : str, optional
        The name of the backend to use. Valid options currently include
        'pickle', 'mongo', 'memory', and 'sql'. If not provided, defaults to
        'pickle' unless the 'mongetter' argument is passed, in which
        case the mongo backend is automatically selected.
    mongetter : callable, optional
        A callable that takes no arguments and returns a pymongo.Collection
        object with writing permissions. If unset a local pickle cache is used
        instead.
    sql_engine : str, Engine, or callable, optional
        SQLAlchemy connection string, Engine, or callable returning an Engine.
        Used for the SQL backend.
    stale_after : datetime.timedelta, optional
        The time delta after which a cached result is considered stale. Calls
        made after the result goes stale will trigger a recalculation of the
        result, but whether a stale or fresh result will be returned is
        determined by the optional next_time argument.
    next_time : bool, optional
        If set to True, a stale result will be returned when finding one, not
        waiting for the calculation of the fresh result to return. Defaults to
        False.
    cache_dir : str, optional
        A fully qualified path to a file directory to be used for cache files.
        The running process must have running permissions to this folder. If
        not provided, a default directory at `~/.cachier/` is used.
    pickle_reload : bool, optional
        If set to True, in-memory cache will be reloaded on each cache read,
        enabling different threads to share cache. Should be set to False for
        faster reads in single-thread programs. Defaults to True.
    separate_files: bool, default False, for Pickle cores only
        Instead of a single cache file per-function, each function's cache is
        split between several files, one for each argument set. This can help
        if you per-function cache files become too large.
    wait_for_calc_timeout: int, optional, for MongoDB only
        The maximum time to wait for an ongoing calculation. When a
        process started to calculate the value setting being_calculated to
        True, any process trying to read the same entry will wait a maximum of
        seconds specified in this parameter. 0 means wait forever.
        Once the timeout expires the calculation will be triggered.
    allow_none: bool, optional
        Allows storing None values in the cache. If False, functions returning
        None will not be cached and are recalculated every call.

    """
    from .config import _global_params

    # Check for deprecated parameters
    if hash_params is not None:
        message = (
            "hash_params will be removed in a future release, "
            "please use hash_func instead"
        )
        warn(message, DeprecationWarning, stacklevel=2)
        hash_func = hash_params
    # Update parameters with defaults if input is None
    backend = _update_with_defaults(backend, "backend")
    mongetter = _update_with_defaults(mongetter, "mongetter")
    # Override the backend parameter if a mongetter is provided.
    if callable(mongetter):
        backend = "mongo"
    core: _BaseCore
    if backend == "pickle":
        core = _PickleCore(
            hash_func=hash_func,
            pickle_reload=pickle_reload,
            cache_dir=cache_dir,
            separate_files=separate_files,
            wait_for_calc_timeout=wait_for_calc_timeout,
        )
    elif backend == "mongo":
        core = _MongoCore(
            hash_func=hash_func,
            mongetter=mongetter,
            wait_for_calc_timeout=wait_for_calc_timeout,
        )
    elif backend == "memory":
        core = _MemoryCore(
            hash_func=hash_func, wait_for_calc_timeout=wait_for_calc_timeout
        )
    elif backend == "sql":
        core = _SQLCore(
            hash_func=hash_func,
            sql_engine=sql_engine,
            wait_for_calc_timeout=wait_for_calc_timeout,
        )
    else:
        raise ValueError("specified an invalid core: %s" % backend)

    def _cachier_decorator(func):
        core.set_func(func)

        @wraps(func)
        def func_wrapper(*args, **kwds):
            nonlocal allow_none
            _allow_none = _update_with_defaults(allow_none, "allow_none", kwds)
            # print('Inside general wrapper for {}.'.format(func.__name__))
            ignore_cache = _pop_kwds_with_deprecation(
                kwds, "ignore_cache", False
            )
            overwrite_cache = _pop_kwds_with_deprecation(
                kwds, "overwrite_cache", False
            )
            verbose = _pop_kwds_with_deprecation(kwds, "verbose_cache", False)
            ignore_cache = kwds.pop("cachier__skip_cache", ignore_cache)
            overwrite_cache = kwds.pop(
                "cachier__overwrite_cache", overwrite_cache
            )
            verbose = kwds.pop("cachier__verbose", verbose)
            _stale_after = _update_with_defaults(
                stale_after, "stale_after", kwds
            )
            _next_time = _update_with_defaults(next_time, "next_time", kwds)
            # merge args expanded as kwargs and the original kwds
            kwargs = _convert_args_kwargs(
                func, _is_method=core.func_is_method, args=args, kwds=kwds
            )

            _print = print if verbose else lambda x: None

            if ignore_cache or not _global_params.caching_enabled:
                return (
                    func(args[0], **kwargs)
                    if core.func_is_method
                    else func(**kwargs)
                )
            key, entry = core.get_entry((), kwargs)
            if overwrite_cache:
                return _calc_entry(core, key, func, args, kwds)
            if entry is None or (
                not entry._completed and not entry._processing
            ):
                _print("No entry found. No current calc. Calling like a boss.")
                return _calc_entry(core, key, func, args, kwds)
            _print("Entry found.")
            if _allow_none or entry.value is not None:
                _print("Cached result found.")
                now = datetime.now()
                if now - entry.time <= _stale_after:
                    _print("And it is fresh!")
                    return entry.value
                _print("But it is stale... :(")
                if entry._processing:
                    if _next_time:
                        _print("Returning stale.")
                        return entry.value  # return stale val
                    _print("Already calc. Waiting on change.")
                    try:
                        return core.wait_on_entry_calc(key)
                    except RecalculationNeeded:
                        return _calc_entry(core, key, func, args, kwds)
                if _next_time:
                    _print("Async calc and return stale")
                    core.mark_entry_being_calculated(key)
                    try:
                        _get_executor().submit(
                            _function_thread, core, key, func, args, kwds
                        )
                    finally:
                        core.mark_entry_not_calculated(key)
                    return entry.value
                _print("Calling decorated function and waiting")
                return _calc_entry(core, key, func, args, kwds)
            if entry._processing:
                _print("No value but being calculated. Waiting.")
                try:
                    return core.wait_on_entry_calc(key)
                except RecalculationNeeded:
                    return _calc_entry(core, key, func, args, kwds)
            _print("No entry found. No current calc. Calling like a boss.")
            return _calc_entry(core, key, func, args, kwds)

        def _clear_cache():
            """Clear the cache."""
            core.clear_cache()

        def _clear_being_calculated():
            """Mark all entries in this cache as not being calculated."""
            core.clear_being_calculated()

        def _cache_dpath():
            """Return the path to the cache dir, if exists; None if not."""
            return getattr(core, "cache_dir", None)

        def _precache_value(*args, value_to_cache, **kwds):  # noqa: D417
            """Add an initial value to the cache.

            Arguments:
            ---------
            value_to_cache : any
                entry to be written into the cache

            """
            # merge args expanded as kwargs and the original kwds
            kwargs = _convert_args_kwargs(
                func, _is_method=core.func_is_method, args=args, kwds=kwds
            )
            return core.precache_value((), kwargs, value_to_cache)

        func_wrapper.clear_cache = _clear_cache
        func_wrapper.clear_being_calculated = _clear_being_calculated
        func_wrapper.cache_dpath = _cache_dpath
        func_wrapper.precache_value = _precache_value
        return func_wrapper

    return _cachier_decorator



"""A MongoDB-based caching core for cachier."""

# This file is part of Cachier.
# https://github.com/python-cachier/cachier

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2016, Shay Palachy <shaypal5@gmail.com>

import pickle  # for serialization of python objects
import sys  # to make sure that pymongo was imported
import time  # to sleep when waiting on Mongo cache\
import warnings  # to warn if pymongo is missing
from contextlib import suppress
from datetime import datetime
from typing import Any, Optional, Tuple

from .._types import HashFunc, Mongetter
from ..config import CacheEntry

with suppress(ImportError):
    from bson.binary import Binary  # to save binary data to mongodb
    from pymongo import ASCENDING, IndexModel
    from pymongo.errors import OperationFailure

from .base import RecalculationNeeded, _BaseCore, _get_func_str

MONGO_SLEEP_DURATION_IN_SEC = 1


class MissingMongetter(ValueError):
    """Thrown when the mongetter keyword argument is missing."""


class _MongoCore(_BaseCore):
    _INDEX_NAME = "func_1_key_1"

    def __init__(
        self,
        hash_func: Optional[HashFunc],
        mongetter: Optional[Mongetter],
        wait_for_calc_timeout: Optional[int],
    ):
        if "pymongo" not in sys.modules:
            warnings.warn(
                "`pymongo` was not found. MongoDB cores will not function.",
                ImportWarning,
                stacklevel=2,
            )  # pragma: no cover

        super().__init__(
            hash_func=hash_func, wait_for_calc_timeout=wait_for_calc_timeout
        )
        if mongetter is None:
            raise MissingMongetter(
                "must specify ``mongetter`` when using the mongo core"
            )
        self.mongetter = mongetter
        self.mongo_collection = self.mongetter()
        index_inf = self.mongo_collection.index_information()
        if _MongoCore._INDEX_NAME not in index_inf:
            func1key1 = IndexModel(
                keys=[("func", ASCENDING), ("key", ASCENDING)],
                name=_MongoCore._INDEX_NAME,
            )
            self.mongo_collection.create_indexes([func1key1])

    @property
    def _func_str(self) -> str:
        return _get_func_str(self.func)

    def get_entry_by_key(self, key: str) -> Tuple[str, Optional[CacheEntry]]:
        res = self.mongo_collection.find_one(
            {"func": self._func_str, "key": key}
        )
        if not res:
            return key, None
        val = None
        if "value" in res:
            val = pickle.loads(res["value"])  # noqa: S301
        entry = CacheEntry(
            value=val,
            time=res.get("time", None),
            stale=res.get("stale", False),
            _processing=res.get("processing", False),
            _completed=res.get("completed", False),
        )
        return key, entry

    def set_entry(self, key: str, func_res: Any) -> None:
        thebytes = pickle.dumps(func_res)
        self.mongo_collection.update_one(
            filter={"func": self._func_str, "key": key},
            update={
                "$set": {
                    "func": self._func_str,
                    "key": key,
                    "value": Binary(thebytes),
                    "time": datetime.now(),
                    "stale": False,
                    "processing": False,
                    "completed": True,
                }
            },
            upsert=True,
        )

    def mark_entry_being_calculated(self, key: str) -> None:
        self.mongo_collection.update_one(
            filter={"func": self._func_str, "key": key},
            update={"$set": {"processing": True}},
            upsert=True,
        )

    def mark_entry_not_calculated(self, key: str) -> None:
        with suppress(OperationFailure):  # don't care in this case
            self.mongo_collection.update_one(
                filter={
                    "func": self._func_str,
                    "key": key,
                },
                update={"$set": {"processing": False}},
                upsert=False,  # should not insert in this case
            )

    def wait_on_entry_calc(self, key: str) -> Any:
        time_spent = 0
        while True:
            time.sleep(MONGO_SLEEP_DURATION_IN_SEC)
            time_spent += MONGO_SLEEP_DURATION_IN_SEC
            key, entry = self.get_entry_by_key(key)
            if entry is None:
                raise RecalculationNeeded()
            if not entry._processing:
                return entry.value
            self.check_calc_timeout(time_spent)

    def clear_cache(self) -> None:
        self.mongo_collection.delete_many(filter={"func": self._func_str})

    def clear_being_calculated(self) -> None:
        self.mongo_collection.update_many(
            filter={
                "func": self._func_str,
                "processing": True,
            },
            update={"$set": {"processing": False}},
        )



import hashlib
import os
import pickle
import threading
from dataclasses import dataclass, replace
from datetime import datetime, timedelta
from typing import Any, Optional, Union

from ._types import Backend, HashFunc, Mongetter


def _default_hash_func(args, kwds):
    # Sort the kwargs to ensure consistent ordering
    sorted_kwargs = sorted(kwds.items())
    # Serialize args and sorted_kwargs using pickle or similar
    serialized = pickle.dumps((args, sorted_kwargs))
    # Create a hash of the serialized data
    return hashlib.sha256(serialized).hexdigest()


@dataclass
class Params:
    """Default definition for cachier parameters."""

    caching_enabled: bool = True
    hash_func: HashFunc = _default_hash_func
    backend: Backend = "pickle"
    mongetter: Optional[Mongetter] = None
    stale_after: timedelta = timedelta.max
    next_time: bool = False
    cache_dir: Union[str, os.PathLike] = "~/.cachier/"
    pickle_reload: bool = True
    separate_files: bool = False
    wait_for_calc_timeout: int = 0
    allow_none: bool = False


_global_params = Params()


@dataclass
class CacheEntry:
    """Data class for cache entries."""

    value: Any
    time: datetime
    stale: bool
    _processing: bool
    _condition: Optional[threading.Condition] = None
    _completed: bool = False


def _update_with_defaults(
    param, name: str, func_kwargs: Optional[dict] = None
):
    import cachier

    if func_kwargs:
        kw_name = f"cachier__{name}"
        if kw_name in func_kwargs:
            return func_kwargs.pop(kw_name)
    if param is None:
        return getattr(cachier.config._global_params, name)
    return param


def set_default_params(**params: Any) -> None:
    """Configure default parameters applicable to all memoized functions."""
    # It is kept for backwards compatibility with desperation warning
    import warnings

    warnings.warn(
        "Called `set_default_params` is deprecated and will be removed."
        " Please use `set_global_params` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    set_global_params(**params)


def set_global_params(**params: Any) -> None:
    """Configure global parameters applicable to all memoized functions.

    This function takes the same keyword parameters as the ones defined in the
    decorator, which can be passed all at once or with multiple calls.
    Parameters given directly to a decorator take precedence over any values
    set by this function.

    Only 'stale_after', 'next_time', and 'wait_for_calc_timeout' can be changed
    after the memoization decorator has been applied. Other parameters will
    only have an effect on decorators applied after this function is run.

    """
    import cachier

    valid_params = {
        k: v
        for k, v in params.items()
        if hasattr(cachier.config._global_params, k)
    }
    cachier.config._global_params = replace(
        cachier.config._global_params,
        **valid_params,  # type: ignore[arg-type]
    )


def get_default_params() -> Params:
    """Get current set of default parameters."""
    # It is kept for backwards compatibility with desperation warning
    import warnings

    warnings.warn(
        "Called `get_default_params` is deprecated and will be removed."
        " Please use `get_global_params` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_global_params()


def get_global_params() -> Params:
    """Get current set of default parameters."""
    import cachier

    return cachier.config._global_params


def enable_caching():
    """Enable caching globally."""
    import cachier

    cachier.config._global_params.caching_enabled = True


def disable_caching():
    """Disable caching globally."""
    import cachier

    cachier.config._global_params.caching_enabled = False
